(* file: lists.ml
   author: Bob Muller

   CS1103 Computer Science I Honors

   Lecture Notes and Code for Meeting 1 of Week 4

  Topics:
    1. A note on World.big_bang
      See the expose.ml code in 03.2.repetitionImages.zip
      found on the Schedule page.
    2. lists
      A. polymorphism
      B. work
*)

[1 + 1; 2 + 2; 3 + 3];;

(* This is an abbreviation for
*)
(1 + 1) :: (2 + 2) :: (3 + 3) :: [];;

(* A list is either:

   - empty, i.e., [] or
   - a cons, i.e., item :: list

   There are no other ways to form lists.
*)

(* add : int list -> int
*)
let rec add ns =
  match ns with
  | [] -> 0
  | m :: ms -> m + add ms

(* length : 'a list -> int
*)
let rec length xs =
  match xs with
  | [] -> 0
  | y :: ys -> 1 + length ys

(* Monomorphism and Polymorphism *******************************

  NB: The use of the + operator in the definition of the add function
  requires the input list to contain integers. Thus, its type is:

   add : int list -> int

  The add function is said to be "monomorphic".

  The definition of the length function doesn't impose any constraints
  at all on what might be in the input list.

   length : 'a list -> int

  The 'a in the type is a -type variable- and the length function is
  said to be "polymorphic".

   append : 'a list -> 'a list -> 'a list
*)
let rec append xs ys =
  match xs with
  | [] -> ys
  | z :: zs -> z :: (append zs ys)

(* NB: How much work will the append function do when called? Let's
  say input xs has N elements and input ys has M elements. Well,
  append calls itself once for each element of xs so it does work
  that is proportional to N. As N grows, so does the amount of work
  done --- we say that append is -linear- in the length of xs.

   reverse : 'a list -> 'a list
*)
let rec reverse xs =
  match xs with
  | [] -> []
  | y :: ys -> append (reverse ys) [y]

(* NB: How much work will this reverse function do when called with an
   input list with N elements? Well, for each element, it calls the
   append function with a list of N - 1 elements. We know that append
   is linear, so reverse requires:

   (N - 1) + (N - 2) + ... + 0

   steps. This is proportional to N(N + 1)/2 = (N^2 + N) / 2. As N
   grows large, the N^2 term dominates both the addition of N and the
   division by 2, so we say that this version of reverse requires work
   proportional to N^2, or is quadratic in the length of xs. If xs has
   1000 elements, this definition will require (10^3)^2 = 10^6 = one
   million steps.
*)
let reverse xs =
  let rec repeat xs answer =
    match xs with
    | [] -> answer
    | y :: ys -> repeat ys (y :: answer)
  in
  repeat xs []

(* This definition of reverse doesn't call append and is much more
   efficient (i.e., better). It requires work proportional to N.
*)

(* Someone asked, is it possible to make a list with different types of
  elements? The answer is yes, but the elements of the list need to be
   included in a sum type. For example, we cannot have a list with either
   ints or floats, but we can make a new sum type number, as follows:
*)
type number = Int of int | Flt of float

(* addNumbers : number list -> float
*)
let rec addNumbers xs =
  match xs with
  | [] -> 0.0
  | (Int n) :: ys -> (float n) +. addNumbers ys
  | (Flt f) :: ys -> f +. addNumbers ys

(* NB: We had to identify which kind of number we matched against and had
   to take the appropriate action on the RHS of the ->.
*)
